# Packages importation
import pandas as pd
import numpy as np
import time
from sklearn.ensemble import RandomForestRegressor


# Class Impute

class Impute:

    def __init__(self, data, names=None,
                 window_size=10, n_trees=50, verbose=False):
        """
        Initialize the class Impute

        Parameters
        ----------
            data : pandas dataframe
                dataframe of interest to be imputed
            names : list of str
                names of the columns to be imputed, defaults to None.
                If None, all columns will be imputed
            windows_size : int
                size of the rolling window used to train the RF, defaults to 10
            n_trees : int
                number of tress in the forest
            verbose : bool
                prints some additionnal informations, defaults to False
        """
        self.__data = data.copy()
        self.__verbose = verbose
        if names is None:
            self.__col_names = list(self.__data.columns)
        else:
            self.__col_names = names
        self.__ws = window_size
        self.__nt = n_trees

        # Add columns with time info
        month = []
        weekday = []
        hour = []
        for i in range(0, self.__data.index.size):
            y, mo, md, h, m, s, wd, yd, idst = self.__data.index[i].timetuple()
            month.append(mo)
            weekday.append(wd)
            hour.append(h)
        self.__data.insert(0, 'Month', month)
        self.__data.insert(1, 'WeekDay', weekday)
        self.__data.insert(2, 'Hour', hour)

        self.countMissing()

    def fillWithRF(self, names=None):
        """
        Fill the columns of the dataframe using the backward-
             forward RF algorithm

        Parameters
        ----------
            names : list of strings
                names of the columns to be imputed, defaults to None.
                If None, all columns of the dataframe are imputed
        """
        names_list = names
        if names_list is None:
            names_list = self.__col_names

        for col_name in names_list:
            rf_f, acc_f = self.trainRF(col_name, forward=True)
            rf_b, acc_b = self.trainRF(col_name, forward=False)
            if self.__verbose:
                print('Imputing variable', col_name,
                      ', OOB accuracy =', 1/2*(acc_f + acc_b))

            start_miss, len_miss = self.info(col_name)

            df_f = self.__data.copy()
            df_b = self.__data.copy()

            for i in range(0, len(start_miss)-1):
                t0 = time.time()

                df_f = self.fill(df_f,
                                 col_name,
                                 rf_f,
                                 start_miss[i],
                                 len_miss[i],
                                 forward=True)
                df_b = self.fill(df_b,
                                 col_name,
                                 rf_b,
                                 start_miss[i],
                                 len_miss[i],
                                 forward=False)

                for j in range(start_miss[i], start_miss[i] + len_miss[i]):
                    if len_miss[i] > 1:
                        f_contrib = (1 - (j-start_miss[i])/(len_miss[i] - 1))*df_f[col_name][j]
                        b_contrib = (j - start_miss[i])/(len_miss[i] - 1)* df_b[col_name][j]
                        self.__data.loc[self.__data.index[j], col_name] = f_contrib + b_contrib
                    else:
                        f_contrib = df_f[col_name][j]
                        b_contrib = df_b[col_name][j]
                        self.__data.loc[self.__data.index[j], col_name] = 1/2*(f_contrib + b_contrib)

                t1 = time.time()

                if self.__verbose:
                    print('  - period', i, 'starting at', start_miss[i],
                          'of length', len_miss[i], ", done in", t1-t0, "s")

    def fill(self, data, col_name, rf, starts, length, forward=True):
        """
        Fill the columns of the dataframe using the RF algorithms

        Parameters
        ----------
            data : pandas dataframe
                Data to be imputed
            col_name : string
                col_name of the column of the dataframe "data" to be imputed.
            rf : sci-kit learn RandomForestRegressor object
                rf object which has been trained over the column "col_name"
                of the dataframe data.
            starts : int
                Starting index of the missing period in the column col_name
                to be imputed
            length : int
                Number of rows of the missing period in the column col_name
                to be imputed.
            forward : bool
                Whether or not the missing period is being imputed
                forward or backward, default to True (forward)
        Returns
        ----------
            data : pandas dataframe
                Dataframe whose missing period starting at index "starts"
                of length "length" has been imputed.
        """
        for j in range(0, length):
            if forward:
                i = starts + j
            else:
                i = starts + length - 1 - j

            prev = []
            prev.append(data['Month'][i])
            prev.append(data['WeekDay'][i])
            prev.append(data['Hour'][i])
            for k in range(0, self.__ws):
                if forward:
                    v = data[col_name][i-1-k]
                else:
                    v = data[col_name][i+1+k]
                prev.append(v)

            if np.isnan(prev).any():
                med = np.nanmedian(prev)
                for k in range(0, len(prev)):
                    if np.isnan(prev[k]):
                        prev[k] = med

            pred = rf.predict([prev])
            data.loc[data.index[i], col_name] = pred

        return data

    def trainRF(self, col_name, forward=True):
        """
        Train a sci-kit learn random forest regressor on the column
        "col_name" of the dataframe data.

        Parameters
        ----------
            col_name : string
                col_name of the column of the dataframe "data" to be imputed.
            forward : bool
                Whether or not the regressor is trained forward or backward,
                default to True (forward).

        Returns
        ----------
            rf : sci-kit learn random forest regressor object
                Trained rf regressor object.
            acc : float
                Out of bag score of the trained random forest "rf"
        """
        Xvals = ['Month', 'WeekDay', 'Hour']
        data = self.__data.loc[:, {col_name,
                                   'WeekDay',
                                   'Hour',
                                   'Month'}].copy()

        # Build the rolling window
        for i in range(1, self.__ws + 1):
            Xvals.append('Xval_t' + str(i))
            if forward:
                data['Xval_t'+str(i)] = data[col_name].shift(i)
            else:
                data['Xval_t'+str(i)] = data[col_name].shift(-i)

        data.dropna(inplace=True)

        # Prepare the data
        X = data[Xvals].values
        Y = data[col_name].values
        rf = RandomForestRegressor(n_estimators=self.__nt,
                                   random_state=42,
                                   oob_score=True,
                                   n_jobs=-1,
                                   warm_start=True,
                                   max_features="sqrt")
        rf.fit(X, Y)
        acc = rf.oob_score_
        return rf, acc

    def countMissing(self):
        """
        Counts the lengths, expressed as a number of rows, of the periods
        of missing and non-missing values.
        The results are stored in a private object's instance attribute e.g
        Impute.__missing where Impute.__missing['M'][col_name]
        (resp: Impute.__missing['NM'][col_name]) contains a list
        of the length of the differents missing (resp: non missing) periods
        in the column 'col_name'.
        """
        self.__info_missing = pd.DataFrame(index=self.__col_names,
                                           columns=['M', 'NM'])

        for col_name in self.__col_names:
            df_temp = self.__data.loc[:, col_name].values
            df_temp_size = df_temp.size
            in_missing = False
            count = 0
            count_nm = 0
            lengths = []
            lengths_nm = []
            for i in range(0, df_temp_size - 1):
                if np.isnan(df_temp[i]):
                    count += 1
                    if not in_missing:
                        in_missing = True
                        lengths_nm.append(count_nm)
                        count_nm = 0
                else:
                    count_nm += 1
                    if in_missing:
                        lengths.append(count)
                        count = 0
                        in_missing = False
            if in_missing:
                lengths.append(count)
            else:
                lengths_nm.append(count_nm)

            self.__info_missing['M'][col_name] = lengths
            self.__info_missing['NM'][col_name] = lengths_nm

    def test(self, col_name, n=1):
        """
        Test the RF algorithm on a window of the column 'col_name'
        of the dataframe which is passed to the Impute class constructor.
        The starting point and the length of the period are for now
        set at random.

        Parameters:
        ----------
            col_name : string
                col_name of the column upon which the RF algorithm will be
                tested
            n : int
                Number of tests performed.

        Returns
        ----------
            pred_values : pandas Time Serie
                The predicted values infered by the trained RF algorithm
                in the missing window.
            true_values : pandas Time Serie
                The true values in the missing window.
        """
        rf_f, acc_f = self.trainRF(col_name, forward=True)
        rf_b, acc_b = self.trainRF(col_name, forward=False)

        print('Test starting for variable ', col_name)
        print('OOB accuracy for this model =', 1/2*(acc_f + acc_b))

        data = self.__data.copy()
        start_miss = 0
        len_miss = 0

        for i in range(0, n):
            start_miss, len_miss, it = self.generateRandomMissing(col_name)
            print('- Test', i + 1, ', starting at', start_miss,
                  ', length =', len_miss)
            data = self.__data.copy()
            # The values of the random window generated
            # in self.generateRandomMissing are set to NaN only here
            # TODO: should be moved to self.generateRandomMissing,
            # more semantically correct
            for k in range(start_miss, start_miss + len_miss):
                data.loc[data.index[k], col_name] = np.nan
            d_f = data.copy()
            d_b = data.copy()
            d_f = self.fill(d_f, col_name, rf_f, start_miss, len_miss,
                            forward=True)
            d_b = self.fill(d_b, col_name, rf_b, start_miss, len_miss,
                            forward=False)
            for j in range(start_miss, start_miss + len_miss):
                f_contrib = (1 - (j - start_miss)/(len_miss - 1))*d_f[col_name][j]
                b_contrib = (j - start_miss)/(len_miss - 1)*d_b[col_name][j]
                data.loc[data.index[j], col_name] = f_contrib + b_contrib

            mse = self.MSE(self.__data[col_name],
                           data[col_name],
                           start_miss,
                           len_miss)
            rsq = 1 - mse/np.var(self.__data[col_name])

            print('  MSE =', mse, 'RSQ =', rsq)
        pred_values = data[col_name][start_miss - self.__ws: start_miss +len_miss + self.__ws]
        true_values = self.__data[col_name][start_miss - self.__ws:start_miss +len_miss + self.__ws]

        return pred_values, true_values

    def test_impute(self, col_name, start_miss, len_miss):
        """
        Test the RF algorithm on a window of the column 'col_name'
        of the dataframe which is passed to the Impute class constructor.
        The starting point and the length of the period are given as arguments.

        Parameters
        ----------
            col_name : string
                col_name of the column upon which the RF algorithm will be
                tested
            start_miss : int
                Starting index of the missing window
            length_miss : int
                Length, expressed as a number of rows, of the missing window
        Returns
        ----------
            pred_values : pandas Time Serie
                The predicted values infered by the trained RF algorithm
                in the missing window.
            true_values : pandas Time Serie
                The true values in the missing window.
            mse: float
                mse value calculated for the imputation algorithm
        """
        rf_f, acc_f = self.trainRF(col_name, forward=True)
        rf_b, acc_b = self.trainRF(col_name, forward=False)

        print('Test starting for variable ', col_name)
        print('OOB accuracy for this model =', 1/2*(acc_f + acc_b))

        data = self.__data.copy()
        for k in range(start_miss, start_miss + len_miss):
            data.loc[data.index[k], col_name] = np.nan

        d_f = data.copy()
        d_b = data.copy()
        d_f = self.fill(d_f, col_name, rf_f, start_miss, len_miss,
                        forward=True)
        d_b = self.fill(d_b, col_name, rf_b, start_miss, len_miss,
                        forward=False)

        for j in range(start_miss, start_miss + len_miss):
            f_contrib = (1 - (j - start_miss)/(len_miss - 1))*d_f[col_name][j]
            b_contrib = (j - start_miss)/(len_miss - 1)*d_b[col_name][j]
            data.loc[data.index[j], col_name] = f_contrib + b_contrib

        mse = self.MSE(self.__data[col_name],
                       data[col_name],
                       start_miss,
                       len_miss)
        rsq = 1 - mse/np.var(self.__data[col_name])

        print('  MSE =', mse, 'RSQ =', rsq)
        pred_values = data[col_name][start_miss - self.__ws: start_miss +
                                     len_miss + self.__ws]
        true_values = self.__data[col_name][start_miss - self.__ws:start_miss +
                                            len_miss + self.__ws]

        return pred_values, true_values, mse

    def generateRandomMissing(self, col_name):
        """
        Generate a window at random on the column 'col_name'
        of the dataframe which is passed to the class constructor.
        WARNING: the window's values are set to NaN only in self.test().
        This behaviour is not coherent and should be fixed in
        the near future.

        Parameters
        ----------
            col_name : string
                name of the column upon which the window
                will be generated

        Returns
        ----------
            start_miss : int
                Starting index of the missing window generated in
                the column 'col_name' of the dataframe passed the constructor
            length_miss : int
                Length, expressed as a number of rows, of the missing window
                generated in the column 'col_name' of the dataframe
                passed to the constructor.
            k : int
                Number of iterations it tooks to generate the missing
                window
        """
        size = self.__data.index.size
        is_ok = False
        k = 0
        while not is_ok:
            k += 1
            len_miss = np.random.randint(10,
                                         np.ceil(self.__data.shape[0]/2), 1)[0]
            start_miss = np.random.randint(self.__ws + 1,
                                           size - self.__ws-len_miss - 1, 1)[0]
            is_ok = not np.isnan(self.__data[col_name][start_miss:start_miss +
                                                       len_miss]).any()

        return start_miss, len_miss, k

    def info(self, col_name):
        """
        Returns the position and length of each missing period in the column
        'col_name' of the dataframe passed to the class constructor

        Parameters
        ----------
            col_name : string
                name of the column whose informations are returned

        Returns
        ----------
            start_miss : numpy array of int
                Contains the starting index of each of the missing periods
                in the column 'col_name' of the dataframe passed
                to the class constructor
            length_miss : numpy array of int
                Contains the lengths, expressed as a number of rows,
                of each of the missing periods of the column 'col_name'
                of the dataframe passed to the class constructor.
        """
        len_miss = self.__info_missing['M'][col_name]
        nm = self.__info_missing['NM'][col_name]
        start_miss = list({nm[0]})
        for i in range(0, len(len_miss)-1):
            start_miss.append(start_miss[i] + len_miss[i] + nm[i+1])
        return start_miss, len_miss

    def printInfo(self, col_name):
        """
        Prints, in a nice formatted way the starting index and the length,
        expressed as a number of rows, of each missing period
        in the column 'col_name'.

        Parameters
        ----------
            col_name : string
                name of the column whose missing periods are printed
        """
        start_miss, len_miss = self.info(col_name)
        print('Variable :', col_name)
        print('Number of missing periods:', len(len_miss))
        for i in range(0, len(len_miss)):
            print('Period', i+1, 'starting at index', start_miss[i],
                  ', length =', len_miss[i])

    def MSE(self, real, pred, starting, length):
        """
        Compute the Mean Squared Error (MSE) between the real and predicted
        values, inferred by the RF backward-forward algorithm,
        on a column upon which the RF is being tested.

        Parameters
        ----------
            real : pandas time serie
                Real values of the missing period in the column tested
            pred : pandas time serie
                Predicted values of the missing period in the column tested
            starting : int
                Starting index of the missing period in the column tested
            length : int
                Length, expressed as a number of rows, of the missing period
                in the column tested

        Returns
        ----------
            mse : real
                MSE computed between real and pred values in the column
                to be tested
        """
        diff = real - pred
        mse = np.mean(np.power(diff[starting:starting + length], 2))
        return mse

    def setWindowSize(self, window_size):
        """
        set the window size upon which the random forest is trained.

        Parameters
        ----------
            window_size : int
            Window size
        """
        self.__ws = window_size

    def setNTrees(self, n_trees):
        """
        set the number of trees in the random forest

        Parameters
        ----------
            n_trees : int
            Number of trees set in the random forest
        """
        self.__nt = n_trees

    def setVerbosity(self, v):
        """
        set whether or not to print additionnal informations to monitor
        the imputing process

        Parameters
        ----------
            v : bool
            If True, some additionnal informations are printed
            during the imputing process
        """
        self.__verbose = v

    def getData(self):
        print(self.__data)
        return self.__data.drop(columns=['Hour', 'Month', 'WeekDay'])

    def getNames(self):
        return self.__col_names

    def getWindowSize(self):
        return self.__ws

    def getNTrees(self):
        return self.nt

    def getMissingTable(self):
        return self.__info_missing

    def getVerbosity(self, v):
        return self.__verbose
